/*  osdchat - on-screen display messaging service
 *  Copyright (c) 2003 by Mike Smelov  <mike-at-argoncorp-dot-com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  See LICENSE for details.
 *
 */

 /* cosdchat.c */

#include "cosdchat.h"
#include "common.h"
#include "version.h"

/* defined in common.c */
extern char *prog_name;

/* print usage information */
static void print_usage()
{
	printf(	"cosdchat - console osdchat client " OC_VERSION "\n"
		"usage:	%s <host> [-D] [-h] [-f <fifo>] [-p <port>]\n\n"
		"\t-D\t\tTurn debugging mode on - only for testing\n"
		"\t-h\t\tUsage information\n"
		"\t-f <file>\tRead input from fifo file\n"
		"\t-p <port>\tSet port\n\n", prog_name);

	exit(EXIT_SUCCESS);
}

/* signal handler function for SIGINT...quit program */
static void leave(int sig)
{
	(void) sig;

	printf("\nQuit.\n");
	tcsetattr(STDIN_FILENO, TCSANOW, &term);
	(void) close(sockfd);
	exit(EXIT_SUCCESS);
}

static void conn_lost()
{
	printf("\nConnection to host was lost.\n");
	tcsetattr(STDIN_FILENO, TCSANOW, &term);
	(void) close(sockfd);
	exit(EXIT_FAILURE);
}

static void sig_alarm(int sig)
{
	int msg = -1;
	int flags;
	(void) sig;

	flags = fcntl(sockfd, F_GETFL);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

	/* read response */
	if ( read(sockfd, &msg, sizeof(int)) < 0 && errno != EAGAIN) {
		perror(NULL);
		close(sockfd);
		exit(EXIT_FAILURE);
	}

	if (msg == MSG_HOSTDIED)
		conn_lost();

	fcntl(sockfd, F_SETFL, flags);
	alarm(5);
}

/* initialize socket and connect to host with timeout */
static void init_connect(char *server_host, int serv_port)
{
	struct hostent *hostinfo;
	struct sockaddr_in addr;
	struct timeval tval;
	fd_set wset;
	int n, flags;
	int sock_err, err_len;

	printf("Connecting to host %s...", server_host);
	(void) fflush(stdout);

	if ( (hostinfo = gethostbyname(server_host)) == NULL)
		printDBG(LVL_FATERR, "%s\n", hstrerror(h_errno));

	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
		printDBG(LVL_FATERR, "socket(): %m\n");

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = *(struct in_addr *)*hostinfo->h_addr_list;
	addr.sin_port = htons((uint16_t) serv_port);

	/* set socket to non-blocking mode */

	flags = fcntl(sockfd, F_GETFL, 0);
	if ( fcntl(sockfd, F_SETFL, flags | O_NONBLOCK) < 0 )
		printDBG(LVL_FATERR, "fcntl(O_NONBLOCK): %m\n");

	if ( connect(sockfd, (struct sockaddr *) &addr, sizeof(addr)) == -1 && errno != EINPROGRESS)
		printDBG(LVL_FATERR, "connect(): %m\n");

	FD_ZERO(&wset);
	FD_SET(sockfd, &wset);
	tval.tv_sec = 4;		/* connection timeout */
	tval.tv_usec = 0;

	n = select(sockfd + 1, NULL, &wset, NULL, &tval);

	if (n == 0)
		printDBG(LVL_FATERR, "%s\n", strerror(ETIMEDOUT));

	if (n < 0)
		printDBG(LVL_FATERR, "select(): %m\n");

	err_len = sizeof(sock_err);
	if ( getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (void *) &sock_err, &err_len) < 0 )
		printDBG(LVL_FATERR, "getsockopt(%d): %m\n", sock_err);

	if (sock_err != 0)
		printDBG(LVL_FATERR, "%s\n", strerror(sock_err));

	if ( fcntl(sockfd, F_SETFL, flags) < 0)
		printDBG(LVL_FATERR, "fcntl(0): %m\n");
}

/* setup the prompt according to given format*/
static int set_prompt(char *format, char *prompt)
{
	struct passwd *pw;
	char host_name[128];
	char tty_name[128];
	char *next;

	printDBG(LVL_DEBUG, "Setting up prompt based on format\n");

	for (next = format; *next != NIL; next++) {

		if (*next != '\\') {
			(void) strncat(prompt, next, 1);
			continue;
		}

		switch(*(++next)) {

		case 's':
			(void) strncat(prompt, " ", 1);
			break;

		case 'h': 	/* get hostname */
			if (gethostname(host_name, 128) < 0) {
				printDBG(LVL_ERROR, "warning: could not get hostname, using 'host'.\n");
				(void) strncat(prompt, "host", 127);
			} else {
				(void) strncat(prompt, host_name, 127);
			}
			break;

		case 'n':	/* get login name */
			if ( (pw = getpwuid(getuid())) == NULL ) {
				printDBG(LVL_ERROR, "warning: could not get user's name, using 'name'.\n");
				(void) strncat(prompt, "name", 127);
			} else {
				(void) strncat(prompt, pw->pw_gecos, 127);
			}
			break;

		case 'u':	/* get login name */
			if ( (pw = getpwuid(getuid())) == NULL ) {
				printDBG(LVL_ERROR, "warning: could not get login name, using 'user'.\n");
				(void) strncat(prompt, "user", 127);
			} else {
				(void) strncat(prompt, pw->pw_name, 127);
			}
			break;

		case 'y':	/* get tty name */
			if (ttyname_r(0, tty_name, 128) != 0) {
				if (pipe_mode) {
					(void) strncat(prompt, "pipe", 127);
				} else {
					printDBG(LVL_ERROR, "warning: could not get tty name, using 'tty'.\n");
					(void) strncat(prompt, "tty", 127);
				}
			} else {
				(void) strncat(prompt, tty_name, 127);
			}
			break;
		}
	}

	printDBG(LVL_DEBUG, "\n");

	return strlen(prompt)+1;
}

/* main event loop */
static void main_loop(void)
{
	char format[BUFSIZE] = {0};
	char prompt[BUFSIZE] = {0};
	char *localmsg;
	char local[BUFSIZ];
	int prompt_len, sent;

	/* save terminal state */
	tcgetattr(STDIN_FILENO, &term);

	/* read prompt format from server */
	(void) read(sockfd, format, BUFSIZE-1);
	printDBG(LVL_DEBUG, "Format: %s (%d)\n", format, strlen(format));

	/* set and write prompt to server */
	prompt_len = set_prompt(format, prompt);
	send(sockfd, prompt, prompt_len, MSG_DONTWAIT);

	if (pipe_mode) {
		while(fgets(local, BUFSIZ, stdin)) {
			usleep(150000);		/* stupid hack - will be removed */
			local[strlen(local)-1] = NIL;
			printDBG(LVL_DEBUG, "local: %s\n", local);
			sent = send(sockfd, local, strlen(local)+1, MSG_DONTWAIT);
			printDBG(LVL_DEBUG, "Sent %d chars\n", sent);
		}
		close(sockfd);
		return;
	}

	printf(	"\ncosdchat " OC_VERSION " - console osdchat client\n" \
		"(leading \"/q\" or ctrl-c to quit)\n\n");

	for (;;) {
		alarm(5);
		localmsg = readline(prompt);

		/* null message */
		if (localmsg[0] == NIL) {
			continue;
		}

		/* quit */
		if (localmsg[0] == '/' && localmsg[1] == 'q') {
			leave(0);
		}

		add_history(localmsg);

		/* send message */
		sent = send(sockfd, localmsg, strlen(localmsg)+1, MSG_DONTWAIT);
		printDBG(LVL_DEBUG, "Sent %d chars\n", sent);

		if (sent == -1)
			printDBG(LVL_DEBUG, "send(): %m\n");

		if (sent != strlen(localmsg)+1)
			printDBG(LVL_DEBUG, "Incomplete message sent to peer, bytes xfer: %d\n", sent);

		free(localmsg);
	}
}

/* ********************************** main ********************************* */
int main(int argc, char **argv)
{
	int opt, fd;
	int serv_port = SERV_PORT;
	msg_t server_msg;
	struct stat statbuf;
	struct servent *sp;

	set_prog_name(argv[0]);

	opterr = 0;
	while ((opt = getopt(argc, argv, "Dh")) != -1) {
		switch(opt) {
		case 'D':
			debug = 1;
			printDBG(LVL_DEBUG, "Debugging ON\n");
			break;
		case 'h':
			print_usage();
		}
	}

	/* not enough args, print usage */
	if (argc < 2)
		print_usage();

	/* program not running on a tty */
	if (!isatty(1)) {
		fprintf(stderr, "Cannot redirect to stdout.\n");
		return EXIT_FAILURE;
	}

	/* input is a pipe not tty */
	if (!isatty(0)) {
		printDBG(LVL_DEBUG, "Input is a pipe.\n");
		pipe_mode = 1;
	}

	optind = 0;
	opterr = 1;
	while ((opt = getopt(argc, argv, "Dhp:f:")) != -1) {
		switch(opt) {
		case 'D':
			break;
		case 'p':
			if ( (serv_port = atoi(optarg)) == 0 ) {
				if ( (sp = getservbyname(optarg, "tcp")) == NULL) {
					printDBG(LVL_SYSLOG, "getservbyname() error for: %s", optarg);
					serv_port = SERV_PORT;
				} else {
					serv_port = ntohs(sp->s_port);
				}
			}
			set_limits(&serv_port, 0, 65535);
			printDBG(LVL_DEBUG, "Setting port to %d\n", serv_port);
			break;
		case 'f':
			/* map specifed fifo file to stdin */
			close(0);

			if ( (fd = open(optarg, O_RDONLY | O_NONBLOCK)) < 0) {
				perror(optarg);
				return EXIT_FAILURE;
			}

			printDBG(LVL_DEBUG, "Reading input from fifo: %s\n", optarg);
			if ( !fstat(fd, &statbuf) ) {
				if ( !S_ISFIFO(statbuf.st_mode) ) {
					printf("Specified file is not a FIFO!\n");
					(void) close(fd);
					return EXIT_FAILURE;
				}
			}
			pipe_mode = 1;
			break;
		case '?':
			print_usage();
		}
	}

	/* hostname not specified */
	if (optind == argc)
		print_usage();

	/* handle signals */
	(void) signal(SIGHUP, leave);
	(void) signal(SIGINT, leave);
	(void) signal(SIGTERM, leave);
	(void) signal(SIGPIPE, conn_lost);
	(void) signal(SIGALRM, sig_alarm);

	/* attempt to initialize socket and connect with timeout */
	init_connect(argv[optind], serv_port);

	/* read returned message from host */
	if ( complete_io(read, sockfd, &server_msg, sizeof(msg_t)) < 0 ) {
		perror(NULL);
		return EXIT_FAILURE;
	}

	printDBG(LVL_DEBUG, "\nMessage: %d\n", server_msg);
	printf("connection established");

	switch(server_msg) {
	case MSG_SERVFULL:
		printf(", but server has reached client limit.\n");
		return MSG_SERVFULL;
	case MSG_NODISP:
		printf(", but no display found.\n");
		return MSG_NODISP;
	case MSG_CLIBLOCK:
		printf(", but you are banned.\n");
		return MSG_CLIBLOCK;
	case MSG_CONNOK:
		break;
	default:
		printf(", but an unknown message was received.\n");
		return -1;
	}
	
	printf(".\n");
	main_loop();

	return EXIT_SUCCESS;
}
